/* Areatag renamer for 2.2+ type packets.
	Solves the duplicate tag problem.
	Optionally retears and strips path and seenby.
	(c) 1995 -2008  Michiel van der Vlist (2:280/5555)     */

#include <stdio.h>
#include <string.h>
#include <dir.h>
#include <dos.h>
#include <stdlib.h>
#include <process.h>
#include <io.h>
#include <fcntl.h>
#include <alloc.h>
#include <time.h>
#include <conio.h>
#include <ctype.h>

typedef unsigned char byte;
typedef unsigned int word;
typedef unsigned long int dword;

#include "gestruct.h"
#include "fmstruct.h"

#define PROGNAME "Marena"
#define VERSION "1.21 beta 20080610"
#define MYNAME "Michiel van der Vlist (2:280/5555)"
#define FTSCCODE 0X18FF
#define TEMP_I "T1"
#define TEMP_O "T2"
#define TRUE (1==1)
#define FALSE (!TRUE)
#define ASCNULL 0
#define CR 0x0D
#define LF 0x0A
#define SR 0x8D
#define IN 0
#define OUT 1

#define RETEAR 1
#define REPATH 2
#define STRIPS 4

#define TOOMANYLINKS 11
#define TOOMANYTAGS 12
#define TOOMANYAKAS 13
#define WRONGNUMBER 14
#define ONOROFF 15
#define BADKEYWORD 16
#define NAMEMISSING 17
#define AREAMISSING 18
#define UNKNOWNTOSSER 19
#define NOGEENVIRON 20
#define NOFMENVIRON 21
#define NOGESETUP 22
#define NOFMCONFIG 23
#define GESETUPERR 24
#define FMCFGERR 25
#define DIRERR 30
#define NOCONFIG 99
#define MEMALLOCERR 100

#define MAXLINKS 50
#define MAXAKAS 20
#define MAXTAGS 500
#define TAGLENG 35

#define CONFIG "MARENA.CFG"
#define DEFAULTLOG "MARENA.LOG"

#define MYDOMAIN "Fidonet"
#define HISDOMAIN "Fidonet"

#define ARC 0
#define ARJ 1
#define LZH 2
#define PAK 3
#define ZIP 4
#define RAR 5

#define STALONE 0
#define UNKNOWN 0
#define GECHO 1
#define FMAIL 2

#define TOSSER0 "UNKNOWN"
#define TOSSER1 "GECHO"
#define TOSSER2 "FMAIL"


typedef unsigned char boolean;
typedef char string[128];

typedef struct {unsigned int zone,net,node,point;} fidonumber;

int do_pkts(int linknr, int inout);
int findpackets(char *indir);
int findpackedfiles(char *fm, char *path);
int do_apacket(int linknr, char *pktn, int inout, int mode);
boolean forusorhim(int *ln, FILE *f, int inout, int *a);
int unpack(char *fn, char *dir, char *dest);
int pack(char *fn, char *dir, char *source, int nr);
int docmd(int cmd, char *cp, char *dp);
int toss(void);
int scan(void);
int calculate(void);
int showconfig(void);
int help(void);
int getoss(char *cp);
int gescan(char *cp);
int fmtoss(char *dp);
int fmscan(char *dp);
int getparagraph(char *s, long n, FILE *f);
void flushparagraph(char *s, FILE *f);
void timestamp(FILE *f, int s);
void pstring(char *st,int n);
void pstring2(char *st, int n, boolean dolog);
void summary(void);
boolean scanfidonumber(char *s, fidonumber *f);
unsigned long calc_crc32(char *name);
void removelf(char *s);
void addslash(char *s);
void removeslash(char *s);
int readconfig(string config, char *p);
void copydirname(char *s, char *t);
void initsome(void);
int isaka(unsigned int z, unsigned int n, unsigned int f, unsigned int p);
void report_error(int err);
int dearchive(char *fn, char *dest);
int get_archive_type(char *fnam);
void renameleftpackets(char *dir);
void logheader(void);
boolean makedirs(void);
int readtosserconfig(int tossertype, int *akacnt);
int readgesetup(int *akacnt);
int readfmailcfg(int *akacnt);

char files_tounpack[500][13];
char pkts_tounpack[500][9];
char logline[255];

const char commandlist[10][8]={"?","HELP","TOSS","SCAN","CALC","SHOW",
										"GETOSS","GESCAN","FMTOSS","FMSCAN"};
const int deflogflag[10]={0,0,1,1,0,0,1,1,1,1};
const char optionlist[6][10]={"-RETEAR","-REPATH","-STRIPS",
															"-LOG","-NOLOG","-VERBOSE"};
const char configlist[][11]={"AKA","INBOUND","SECURE","OUTBOUND"};
boolean retear=FALSE,repath=FALSE,stripseenby=FALSE,logisopen=FALSE;
int logflag=0;
FILE *logf;
unsigned int nrpackets=0, nrmessages=0, nrrenamed=0;
unsigned int nrrepathed=0, nrreteared=0, nrsstripped=0;
int clinenr=0, Tosserid=0, Verbose=0, FtscprodcodeOn=0;
int Versionmajor,Versionminor;
boolean Fm10x;

fidonumber Aka[MAXAKAS],links[MAXLINKS];
char Hisfiles[MAXLINKS][9];
char Myfiles[MAXLINKS][9];
int Tagpntrs[MAXLINKS],Tagaantal[MAXLINKS];
char Histags[MAXTAGS][TAGLENG];
char Mytags[MAXTAGS][TAGLENG];
byte Flags[MAXLINKS];
string Logfile,Inbound,Secure,Outbound,Packets,Pkttemp,Pkttemp_i,Pkttemp_o;


int main (int argc, char *argv[])
{
	int i,j,k,n=0,r=0;
	char command[sizeof(commandlist[0])],*cp;

	initsome();						/* init some global variables */
	printf("%s %s Echotag renamer (c) %s \n",PROGNAME,VERSION,MYNAME);

	for (i=1;i<argc;i++)			/* then scan the options */
	{
		for (j=0;j<sizeof(optionlist)/sizeof(optionlist[0]);j++)
		{
			if (strnicmp(optionlist[j],argv[i],4)==0)
			{
				switch(j)
				{
					case 0:  retear=TRUE; break;
					case 1:	repath=TRUE; break;
					case 2:	stripseenby=TRUE;  break;
					case 3:	logflag++; break;
					case 4:	logflag--; break;
					case 5:	Verbose++; break;
				}
			}
		}
	}
	if ((i=readconfig(CONFIG,argv[0]))!=0)	/* then read the config */
	{
		report_error(i); exit(i);
	}

	strcpy(logline,"Command line:");
	for (i=1; i<argc; i++) { strcat(logline," "); strcat(logline,argv[i]); }
	strcat(logline,"\n");
	pstring(logline,0);

	if (!makedirs()) exit(DIRERR);

	for (i=1;i<argc;i++)       /* Then scan the commands */
	{
		cp=NULL;
		for (j=0;j<sizeof(command)-1;j++)
		{
			if (argv[i][j]==0) break;
			if (argv[i][j]=='=')	{ cp=&argv[i][j+1]; break; }
			command[j]=argv[i][j];
		}
		command[j]=0;

		for (j=0;j<sizeof(commandlist)/sizeof(commandlist[0]);j++)
		{
			if (strnicmp(commandlist[j],command,6)==0)
			{
				logflag=logflag+deflogflag[j];
				logheader();
				k=docmd(j,cp,argv[argc-1]);
				logflag=logflag-deflogflag[j];
				n++;
				if (k<0) { r=0-k; break; }
				r=max(r,k);
			}
		}
	}
	if (n==0) help();
	if (logisopen) { logisopen=FALSE; logflag=-128; fclose(logf); }
	return(r);
}

void logheader(void)
{
	struct tm *local;
	time_t l;
	const char *monthn[]={"***","Jan","Feb","Mar","Apr","May","Jun",
									"Jul","Aug","Sep","Oct","Nov","Dec"};
	const char *dow[] = {"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};

	if (logisopen || logflag<=0) return;

	if ((logf=fopen(Logfile,"at"))!=NULL)
	{
		time(&l);
		local = localtime(&l);
		fprintf(logf,"\n----------  %s %02d %s %04d, %s v %s\n",
		dow[local->tm_wday], local->tm_mday, monthn[local->tm_mon + 1],
											(local->tm_year+1900),PROGNAME,VERSION);
		logisopen=TRUE;
	}
	else logflag=0;
}

int docmd(int cmd, char *cp, char *dp)
{
	switch (cmd)
	{
		case 0:
		case 1: return(help());
		case 2: return(toss());
		case 3: return(scan());
		case 4: return(calculate());
		case 5: return(showconfig());
		case 6: return(getoss(cp));
		case 7: return(gescan(cp));
		case 8: return(fmtoss(dp));
		case 9: return(fmscan(dp));
	}
	return(0);
}

int toss(void)
{
	int i,j,n;
	static int didit=0;

	if (didit==0)
	{
		printf("Tossing..\n");
		j=0;
		while (j<MAXLINKS && Hisfiles[j][0])
		{
			n=findpackedfiles(Hisfiles[j],Inbound);
			for (i=0;i<n;i++)
			{
				if (unpack(files_tounpack[i],Inbound,Pkttemp)==0) do_pkts(j,IN);
			}
			if (Secure[0])
			{
				n=findpackedfiles(Hisfiles[j],Secure);
				for (i=0;i<n;i++)
				{
					if (unpack(files_tounpack[i],Secure,Pkttemp)==0) do_pkts(j,IN);
				}
			}
			j++;
		}
		didit++;
		summary();
	}
	return(0);
}

int scan(void)
{
	int i,j,n,m;
	static int didit=0;

	if (didit==0)
	{
		printf("Scanning...\n");
		j=0;
		while (j<MAXLINKS && Myfiles[j][0])
		{
			renameleftpackets(Pkttemp_i);
			n=findpackedfiles(Myfiles[j],Outbound);
			for (i=0;i<n;i++)
			{
				if (unpack(files_tounpack[i],Outbound,Pkttemp_i)==0)
				{
					m=do_pkts(j,OUT);
					pack(files_tounpack[i],Outbound,Pkttemp_o,m);
				}
			}
			j++;
		}
		didit++;
		summary();
	}
	return(0);
}

int getoss(char *cp)
{
	int r=0;

	r=do_apacket(-1,cp,IN,GECHO);
	summary();
	return(r);
}

int gescan(char *cp)
{


	return(*cp);
}

int fmtoss(char *dp)
{
	int i,n,r=0,s;
	string pkt;

	if ((n=findpackets(dp))!=0)
	{
		for (i=0;i<n;i++)
		{
			strcpy(pkt,dp);
			strcat(pkt,pkts_tounpack[i]);
			strcat(pkt,".PKT");
			s=do_apacket(-1,pkt,IN,FMAIL);
			r=max(r,s);
		}
	}
	summary();
	return(0-r);
}

int fmscan(char *dp)
{
	int r=0;

	r=do_apacket(-1,dp,OUT,FMAIL);

	summary();
	return(0-r);
}

int showconfig(void)
{
	int i,j,k,l,n;
	char nbuf[20], onoff[4]="off";

	char on[]={"on"};
	char off[]={"off"};
	char tossers[][11]={TOSSER0,TOSSER1,TOSSER2};
	char fmver[10];

	char *pretear, *prepath, *pstrips;

	sprintf(logline,"General:\n");
	pstring(logline,0);
	sprintf(logline,"Verbosity : %d\n",Verbose);
	pstring(logline,0);
	if (FtscprodcodeOn) sprintf(onoff,"on");
	sprintf(logline,"FTSC product code (%04X): %s\n",FTSCCODE,onoff);
	pstring(logline,0);
	if (Tosserid)
	{
		if (Tosserid==FMAIL)
		{
			if (Fm10x) strcpy(fmver,"1.0x"); else strcpy(fmver,">=1.20");
			sprintf(logline,"Tosser    : %s %s\n",tossers[Tosserid],fmver);
		}
		else sprintf(logline,"Tosser    : %s\n",tossers[Tosserid]);
		pstring(logline,0);
	}
	n=sprintf(logline,"Your AKA's: ");
	i=0;
	while (Aka[i].zone && i<MAXAKAS)
	{
		while (n%17) { strcat(logline," "); n++; }
		if (n>55) { strcat(logline,"\n"); pstring(logline,0); n=0; logline[0]=0;}
		n=n+sprintf(nbuf,"%u:%u/%u.%u ",Aka[i].zone,Aka[i].net,Aka[i].node,Aka[i].point);
		strcat(logline,nbuf);
		i++;
	}
	strcat(logline,"\n");
	pstring(logline,0);
	sprintf(logline,"Logfile             : %s\n",Logfile);
	pstring(logline,0);
	sprintf(logline,"Inbound files       : %s\n",Inbound);
	pstring(logline,0);
	sprintf(logline,"Secure inbound files: %s\n",Secure);
	pstring(logline,0);
	sprintf(logline,"Outbound files      : %s\n",Outbound);
	pstring(logline,0);
	sprintf(logline,"Tossers PKT dir.    : %s\n",Packets);
	pstring(logline,0);
	sprintf(logline,"Temp. toss PKT dir. : %s\n",Pkttemp);
	pstring(logline,0);
	sprintf(logline,"Temp1 scan PKT dir. : %s\n",Pkttemp_i);
	pstring(logline,0);
	sprintf(logline,"Temp2 scan PKT dir. : %s\n",Pkttemp_o);
	pstring(logline,0);
	i=0;
	while (links[i].zone && i<MAXLINKS)
	{
		logline[0]=0;
		for (j=0;j<50;j++) strcat(logline,"-");
		strcat(logline,"\n");
		pstring(logline,0);
		sprintf(logline,"Link to node %u:%u/%u.%u\n",links[i].zone,links[i].net,
															links[i].node,links[i].point);
		pstring(logline,0);
		if (Hisfiles[i][0])
		{								  /* only display if set */
			sprintf(logline,"  His bundle name: %s\n",Hisfiles[i]);
			pstring(logline,0);
		}
		if (Myfiles[i][0])
		{
			sprintf(logline,"  My bundle name : %s\n",Myfiles[i]);
			pstring(logline,0);
		}
		if (Flags[i] & RETEAR) pretear=on; else pretear=off;
		if (Flags[i] & REPATH) prepath=on; else prepath=off;
		if (Flags[i] & STRIPS) pstrips=on; else pstrips=off;
		sprintf(logline,"  Retear=%s Repath=%s Strips(eenby)=%s\n",pretear,
																				prepath,pstrips);
		pstring(logline,0);
		k=Tagpntrs[i]; l=0;
		while (Histags[k][0] && k<MAXTAGS && l<Tagaantal[i])
		{
			sprintf(logline,"  Retag '%s' <--> '%s'\n",Histags[k],Mytags[k]);
			pstring(logline,0);
			k++; l++;
		}
		i++;
	}
	printf("\n");
	return(0);
}

void initsome(void)
{
	int i,j;

	for (i=0;i<MAXAKAS;i++) Aka[i].zone=0;
	for (i=0;i<MAXLINKS;i++)
	{
		links[i].zone=0;
		Hisfiles[i][0]=0;	Myfiles[i][0]=0;
		Tagpntrs[i]=0; Tagaantal[i]=0; Flags[i]=0;
	}
	for (j=0;j<MAXTAGS;j++) { Histags[j][0]=0; Mytags[j][0]=0;}
	strcpy(Logfile,DEFAULTLOG);
	sscanf(VERSION,"%d.%d",&Versionmajor,&Versionminor);
}

int help(void)
{
	printf("usage: Marena <commands> <options>\n");
	printf("Commands:\n");
	printf("    toss      -  process inbound files\n");
	printf("    scan      -  process outbound files\n");
	printf("    calc      -  calculate ARCmail file name\n");
	printf("    show      -  show config as read from %s.CFG\n",PROGNAME);
	printf("    getoss=@  -  process inbound pkt in Gecho mode\n");
	printf("    fmtoss    -  process inbound pkt's in Fmail mode\n");
	printf("    fmscan    -  process outbound pkt's in Fmail mode\n");
	printf("    help or ? -  display this message\n");
	printf("Options:\n");
	printf("   -log       -  enable logging to %s if disabled\n",Logfile);
	printf("   -nolog     -  disable logging to %s if enabled\n",Logfile);
	printf("   -verbose   -  enable verbose log mode\n");
	printf("   -retear    -  replace tearline with --- %s %s\n",PROGNAME,VERSION);
	printf("   -repath    -  remove path and add @SPTH %s#%u:%u/%u.%u\n"
													,MYDOMAIN,Aka[0].zone,Aka[0].net,Aka[0].node,Aka[0].point);
	printf("   -strips    -  remove seen-by\n");

	return(0);
}

void summary(void)
{
	printf("\n");
	sprintf(logline,"Packets: %u  Messages: %u  Tags renamed: %u\n",nrpackets,nrmessages,nrrenamed);
	pstring(logline,0);
	sprintf(logline,"Repathed: %u  Reteared: %u  Seenby's stripped: %u\n",nrrepathed,nrreteared,nrsstripped);
	pstring(logline,0);
}

int unpack(char *fn, char *dir, char *dest)
{
	char fname[128];
	int er=-1,fh;
	long fleng=0;

	strcpy(fname,dir);
	strcat(fname,fn);
	if  ((fh=open(fname,O_RDONLY))!=-1);
	{
		fleng=filelength(fh);
		close(fh);
		if (fleng>0)		/* skip zero length files */
		{
			sprintf(logline,"Unpacking %s\n",fname);
			pstring(logline,0);
			er=dearchive(fname,dest);

			if (er==0)	{ remove(fname);	printf("%s removed\n",fname); }
			if (er==-1)
			{
				sprintf(logline,"Problem executing unarchive program, aborting\n");
				pstring(logline,1);
				exit(10);
			}
		}
	}
	return(er);
}

int dearchive(char *fn, char *dest)
{
	int type, err=1;

	type=get_archive_type(fn);
	if (type==-1)
	{
		sprintf(logline,"%s is not ZIP, ARJ, LZH, RAR or PAK/ARC; aborting\n",fn);
		pstring(logline,1);
		exit(11);
	}

	switch (type)
	{
		case ZIP:	err=spawnlp(P_WAIT,"PKUNZIP","PKUNZIP","-o",fn,dest,NULL);
						break;
		case ARJ:	err=spawnlp(P_WAIT,"ARJ","ARJ","e","-y",fn,dest,NULL);
						break;
		case LZH:	err=spawnlp(P_WAIT,"LHA","LHA","e",fn,dest,NULL);
						break;
		case ARC:
		case PAK:	err=spawnlp(P_WAIT,"PAK","PAK","e",fn,dest,NULL);
						break;
		case RAR:	err=spawnlp(P_WAIT,"UNRAR","UNRAR","e","-o-","-c-",fn,dest,NULL);
						break;
	}

	return(err);
}

int get_archive_type(char *fnam)
{
	FILE *f;
	byte bf[11];
	int i;

	if ((f=fopen(fnam,"rb"))==NULL) return(-1);

	for (i=0;i<sizeof(bf)-1;i++) bf[i]=fgetc(f);
	fclose(f);

	if (bf[0]=='R' && bf[1]=='a' && bf[2]=='r' && bf[3]=='!') return(RAR);
	if (bf[0]=='P' && bf[1]=='K') return(ZIP);
	if (bf[0]==0x60 && bf[1]==0xEA) return(ARJ);
	if (bf[3]=='l' && bf[4]=='h') return(LZH);
	if (bf[0]==0x1A) return(PAK);
	return(-1);
}


int pack(char *fn, char *dir, char *source, int nr)
{
	char fname[128], pktname[128] ,ext[4];
	int i,er;
	static int ren=0;

	strcpy(fname,dir);
	strcat(fname,fn);
	strcpy(pktname,source);
	strcat(pktname,"*.PKT");
	sprintf(logline,"Packing %s\n",fname);
	pstring(logline,0);
	er=spawnlp(P_WAIT,"PKZIP","PKZIP",fname,pktname,NULL);
	if (er==-1)
	{
		sprintf(logline,"Problem executing PKZIP, aborting\n");
		pstring(logline,1);
		exit(10);
	}
	if (er>0)
	{
		sprintf(ext,"P%02d",ren);
		sprintf(logline,"PKZIP reports errorlevel %d, aborting\n",er);
		pstring(logline,1);
		sprintf(logline,"All .PKT's are renamed to %s\n",ext);
		pstring(logline,1);
		sprintf(logline,"Please report this to the author, thank you.\n");
		pstring(logline,1);
		for (i=0;i<nr;i++)
		{
			strcpy(pktname,source);					strcpy(fname,source);
			strcat(pktname,pkts_tounpack[i]);   strcat(fname,pkts_tounpack[i]);
			strcat(pktname,".PKT");             strcat(fname,ext);
			rename(pktname,fname);
		}
		ren++; ren=ren%100;
	}

	if (er==0)
	{
		for (i=0;i<nr;i++)
		{
			strcpy(pktname,source);
			strcat(pktname,pkts_tounpack[i]);
			strcat(pktname,".PKT");
			remove(pktname);
		}
	}
	return(er);
}

int do_pkts(int linknr, int inout)
{
	int i,n=0;
	char *t;

	if (inout==IN) t=Pkttemp; else t=Pkttemp_i;

	if ((n=findpackets(t))!=0)
	{
		for (i=0;i<n;i++) do_apacket(linknr,pkts_tounpack[i],inout,STALONE);
	}
	return(n);
}

int findpackedfiles(char *fm, char *path)
{
	struct ffblk infoblk;
	char far *dtasave;
	char match[128], ext[5];
	int j,n=0;
	const char dayn[7][3]={"MO","TU","WE","TH","FR","SA","SU"};

	for (j=0;j<sizeof(dayn)/sizeof(dayn[0]);j++)
	{
		strcpy(match,path);
		strcat(match,fm);
		strcat(match,".");
		strcat(match,dayn[j]);
		strcat(match,"?");
		dtasave=getdta();				/* Save het org. disk transfer adres. */
		if (findfirst(match,&infoblk,0)==0)
		{
			fnsplit(infoblk.ff_name,NULL,NULL,files_tounpack[n],ext);
			strcat(files_tounpack[n++],ext);
			while (findnext(&infoblk)==0)
			{
				fnsplit(infoblk.ff_name,NULL,NULL,files_tounpack[n],ext);
				strcat(files_tounpack[n++],ext);
			}
		}
		setdta(dtasave);              /* org disk transfer adres herstellen */
	}
	return(n);
}

int findpackets(char *indir)
{
	struct ffblk infoblk;
	char far *dtasave;
	char match[128];
	int n=0;

	strcpy(match,indir);
	strcat(match,"????????");
	strcat(match,".PKT");

	dtasave = getdta();				/* Save het org. disk transfer adres. */
	if (findfirst(match,&infoblk,0)==0)
	{
		fnsplit(infoblk.ff_name,NULL,NULL,pkts_tounpack[n++],NULL);
		while (findnext(&infoblk)==0)
		{
			fnsplit(infoblk.ff_name,NULL,NULL,pkts_tounpack[n++],NULL);
		}
	}
	setdta(dtasave);              /* org disk transfer adres herstellen */
	return(n);
}

int do_apacket(int linknr, char *pktn, int inout, int mode)
{
	char inpkt[128],outpkt[128], anotherbuf[64], pktheader[0x3C];
	FILE *in, *out;
	char far *buffer;
	int a,i,b,c,n,g,linecount,tag,ln;
	unsigned int number;
	long bufsize;
	boolean seenorigin,seenspath,madepath,donesome=FALSE;

	ln=linknr;

	if ((bufsize=farcoreleft()-5000)>65768L) bufsize=65768L;

	buffer=farmalloc(bufsize);
	if (buffer==NULL || bufsize<32767L)
	{
		sprintf(logline,"Memory allocation error! Aborting now...\n\a");
		pstring(logline,0);
		if (logisopen) { logisopen=FALSE; logflag=-128; fclose(logf); }
		exit(10);
	}
	if (mode==STALONE)
	{
		if (inout==IN)
		{
			strcpy(inpkt,Pkttemp);
			strcpy(outpkt,Packets);
		}
		else
		{
			strcpy(inpkt,Pkttemp_i);
			strcpy(outpkt,Pkttemp_o);
		}
		strcat(inpkt,pktn);
		strcat(inpkt,".PKT");
		strcat(outpkt,pktn);
		strcat(outpkt,".PKT");
	}
	else
	{
		strcpy(inpkt,pktn);
		for (i=0;i<strlen(pktn);i++)
		{
			outpkt[i]=pktn[i];
			if (pktn[i]=='.') break;
		}
		outpkt[i+1]=0;
		strcat(outpkt,"$TM");
	}
	sprintf(logline,"%s ",inpkt);
	pstring(logline,0);

	if ((in=fopen(inpkt,"rb"))==NULL) return(1);

	if (mode==STALONE) out=fopen(outpkt,"wb");

	if (forusorhim(&ln,in,inout,&a))
	{
		donesome=TRUE;
		nrpackets++;
		if (mode!=STALONE) out=fopen(outpkt,"wb");
		for (i=0;i<0x3C;i++) pktheader[i]=fgetc(in); /* read packet header */
		if (inout==OUT && FtscprodcodeOn)
		{
			pktheader[0x18]=FTSCCODE & 0xFF; /* insert our FTSC prodcut code */
			pktheader[0x2A]=(FTSCCODE >> 8) & 0xFF;
			pktheader[0x19]=Versionmajor & 0xFF; /* insert our version number */
			pktheader[0x2B]=Versionminor & 0xFF;

		}
		for (i=0;i<0x3C;i++) fputc(pktheader[i],out);  /* write packet header */
		n=0;
		do
		{
			nrmessages++;
			for (i=0;i<0x20;i++) fputc(fgetc(in),out);	/* copy header & date */
			do
			{
				fputc((c=fgetc(in)),out);
			}	while (c!=EOF && c!=0);	/* copy tousername */
			do
			{
				fputc((c=fgetc(in)),out);
			}	while (c!=EOF && c!=0);	/* copy fromusername */
			do
			{
				fputc((c=fgetc(in)),out);
			}	while (c!=EOF && c!=0);	/* copy subject */
			i=0;

			/* Ok, we got the header, now go scan the message content */

			linecount=0;
			seenorigin=seenspath=madepath=FALSE;

			while ((g=getparagraph(buffer,bufsize-1,in))!=ASCNULL && g!=EOF)
			{
				if (buffer[0]==0x01)			/* a kludge line? */
				{
					if (inout==OUT)
					{
						if (strnicmp(&buffer[1],"PATH:",5)==0 && (Flags[ln] & REPATH)) continue;
						if (strnicmp(&buffer[1],"SPTH:",5)==0)
						{
							seenspath=TRUE;
							if (madepath) continue;
						}
					}
					flushparagraph(buffer,out);
					continue;
				}
				linecount++;			/* count visible lines only */

				if ((linecount==1) && (strnicmp(buffer,"AREA:",5)==0))
				{
					sprintf(logline,"Msg nr %d %s ",n,&buffer[5]);
					if (inout==IN)
					{
						for (tag=0;tag<Tagaantal[ln];tag++)
						{
							if (stricmp(&buffer[5],Histags[Tagpntrs[ln]+tag])==0)
							{
								fprintf(out,"AREA:%s\r",Mytags[Tagpntrs[ln]+tag]);
								sprintf(anotherbuf," ---> %s\n",Mytags[Tagpntrs[ln]+tag]);
								nrrenamed++;
								strcat(logline,anotherbuf);
								break;
							}
						}
						if (tag==Tagaantal[ln])
						{
							flushparagraph(buffer,out);
							strcat(logline," Tag unchanged\n");
						}
						pstring2(logline,2,Verbose>0);
						continue;
					}
					else
					{
						for (tag=0;tag<Tagaantal[ln];tag++)
						{
							if (stricmp(&buffer[5],Mytags[Tagpntrs[ln]+tag])==0)
							{
								fprintf(out,"AREA:%s\r",Histags[Tagpntrs[ln]+tag]);
								sprintf(anotherbuf," ---> %s\n",Histags[Tagpntrs[ln]+tag]);
								nrrenamed++;
								strcat(logline,anotherbuf);
								if ((Flags[ln] & REPATH) && !seenspath)
								{
									fprintf(out,"%cSPTH: %s#%u:%u/%u.%u\r"
													,1,MYDOMAIN,Aka[a].zone,Aka[a].net,Aka[a].node,Aka[a].point);
									madepath=TRUE;
								}
								break;
							}
						}
						if (tag==Tagaantal[ln])
						{
							flushparagraph(buffer,out);
							strcat(logline," Tag unchanged\n");
						}
						pstring2(logline,2,Verbose>0);
					}
					continue;
				}
				if ((linecount>1) && (!seenorigin) && (strncmp(buffer,"---",3)==0)
													&& (isspace(buffer[3]) || buffer[3]==0))
				{
					if (inout==OUT && (Flags[ln] & RETEAR))
					{
						fprintf(out,"--- %s %s\r",PROGNAME,VERSION);
						nrreteared++;
					}
					else flushparagraph(buffer,out);
					continue;
				}
				if ((linecount>1) && (strncmp(buffer," * Origin: ",11)==0))
				{
					seenorigin=TRUE;
					flushparagraph(buffer,out);
					continue;
				}
				if (inout==OUT && (Flags[ln] & STRIPS) && seenorigin && strncmp(buffer,"SEEN-BY:",8)==0)
				{
					nrsstripped++;
					continue;
				}
				flushparagraph(buffer,out);
			}
			fputc(0,out);
			b=fgetc(in); fputc(b,out);
			c=fgetc(in); fputc(c,out);
			if (b==EOF || c==EOF) number=0; else number=(c<<8)+b;
			if ((Flags[ln] & REPATH) && (seenspath || madepath)) nrrepathed++;
			n++;
		}  while (number);		/* next message  in packet if any */
		fclose(out);
	}
	else
	{
		if (mode==STALONE)
		{
			while ((b=fgetc(in))!=EOF) fputc(b,out);	/* copy rest of packet */
			fclose(out);
		}
	}
	farfree(buffer);
	fclose(in);
//	remove(inpkt);
	if (mode==STALONE)
	{
		remove(inpkt);
		sprintf(logline,"Written to %s\n",outpkt);
		pstring(logline,0);
	}
	else
	{
		if (donesome) { remove(inpkt); rename(outpkt,inpkt); }
	}
	return(0);
}

/* read a paragraph into the buffer, sort of like fgets, but for
	very long lines */

int getparagraph(char *s, long n, FILE *f)
{
	int c;
	size_t i=0;

	while (((c=fgetc(f))!=EOF) && (c!=CR) && (c!=ASCNULL) && (i<n-1))
	{
		if (c==LF) continue;

		s[i]=c;
		i++;
	}
	s[i]=0;
	return(c);
}

void flushparagraph(char *s, FILE *f)
{
	unsigned int i=0;

	while (s[i]) fputc(s[i++],f);
	fputc('\r',f);
}


/* scan the packet header and see if it is for us or him */

boolean forusorhim(int *ln, FILE *f, int inout, int *a)
{
	unsigned int orgzone,destzone,orgnet,destnet,
											orgnode,destnode,orgpoint,destpoint;
	int i;

	fseek(f,0L,SEEK_SET);
	orgnode=fgetc(f)+(fgetc(f)<<8);
	destnode=fgetc(f)+(fgetc(f)<<8);
	fseek(f,0x14L,SEEK_SET);
	orgnet=fgetc(f)+(fgetc(f)<<8);
	destnet=fgetc(f)+(fgetc(f)<<8);
	fseek(f,0x2EL,SEEK_SET);
	orgzone=fgetc(f)+(fgetc(f)<<8);
	destzone=fgetc(f)+(fgetc(f)<<8);
	orgpoint=fgetc(f)+(fgetc(f)<<8);
	destpoint=fgetc(f)+(fgetc(f)<<8);
	rewind(f);
	sprintf(logline,"from %u:%u/%u.%u to %u:%u/%u.%u\n",orgzone,orgnet,orgnode,
									orgpoint,destzone,destnet,destnode,destpoint);
	pstring(logline,-1);

	if (inout==IN)
	{
		if (*ln<0)
		{
			for (i=0;i<MAXLINKS;i++)
			{
				if (links[i].zone==orgzone && links[i].net==orgnet &&
					links[i].node==orgnode && links[i].point==orgpoint)
				{
					*ln=i; break;
				}
			}
			if (*ln<0) return(FALSE);
		}
		return(links[*ln].zone==orgzone && links[*ln].net==orgnet &&
				links[*ln].node==orgnode && links[*ln].point==orgpoint &&
				(*a=isaka(destzone,destnet,destnode,destpoint))>=0);
	}
	else
	{
		if (*ln<0)
		{
			for (i=0;i<MAXLINKS;i++)
			{
				if (links[i].zone==destzone && links[i].net==destnet &&
					links[i].node==destnode && links[i].point==destpoint)
				{
					*ln=i; break;
				}
			}
			if (*ln<0) return(FALSE);
		}
		return(links[*ln].zone==destzone && links[*ln].net==destnet &&
				links[*ln].node==destnode && links[*ln].point==destpoint &&
				(*a=isaka(orgzone,orgnet,orgnode,orgpoint))>=0);
	}
}

/* timestamp to log file */

void timestamp(FILE *f, int s)
{
	const char ind[]={'#','!','?','$','*',':','+','-','%','~','='};
	time_t l;
	struct tm *loc;

	time(&l);
	loc=localtime(&l);
	fprintf(f,"%c %2d:%02d:%02d  ",ind[s],loc->tm_hour,loc->tm_min,loc->tm_sec);
}

/* string output to screen and to log */

void pstring(char *st,int n)
{
	printf("%s",st);
	if (logflag>0)
	{
		if (n>=0) timestamp(logf,n);
		fprintf(logf,"%s",st);
	}
}

/* String to output and conditionally to log */

void pstring2(char *st, int n, boolean dolog)
{
	printf("%s",st);
	if (logflag>0 && dolog)
	{
		if (n>=0) timestamp(logf,n);
		fprintf(logf,"%s",st);
	}
}



/* calculate the name if the xxxxxxxx.ddx file that are used by the
	tosser to send compressed mail packets.

	The ARCmail .060 convention uses:
	((orignet-destnet) << 16) + (orignode-destnode)

	These are also used by Gecho and Fmail if the ARCmail compatibility
	switch in the nodelist manager is on or if it is mail to and from the
	zoneof the main AKA.

	For 4D points Gecho and Fmail calculate a CRC-32 over a string z:n/f.p

	z = DestZone
	n = Aka[0].point + DestNet
	f = DestNode - Aka[0].net
	p = Aka[0].node + DestPoint		*/


int calculate(void)
{
	unsigned long crc32;
	char namebuf[35];
	fidonumber orig,dest;

	printf("Calculate the filename used by Tossers.\n");
	do
	{
		while (printf("\nOrigin node address? "),scanf("%s",namebuf),
						scanfidonumber(namebuf,&orig)==FALSE)
		{
			printf("\n%s is not a valid address!",namebuf);
		}
		while (printf("\nDestination node address? "),scanf("%s",namebuf),
						scanfidonumber(namebuf,&dest)==FALSE)
		{
			printf("\n%s is not a valid address!",namebuf);
		}
		sprintf(namebuf,"%u:%u/%u.%u",dest.zone,orig.point+dest.net,
												dest.node-orig.net,orig.node+dest.point);
		crc32=calc_crc32(namebuf);
		printf("\n");
		sprintf(logline,"Orignode: %u:%u/%u.%u  Destnode: %u:%u/%u.%u\n",
							orig.zone,orig.net,orig.node,orig.point,dest.zone,dest.net,dest.node,dest.point);
		pstring(logline,0);
		sprintf(logline,"ARCmail 0.60   Gecho/Fmail\n");
		pstring(logline,0);
		sprintf(logline,"%04X%04X       %04X%04X\n",orig.net-dest.net,orig.node-dest.node,
						(unsigned)(crc32>>16),(unsigned)crc32&0xFFFF,namebuf);
		pstring(logline,0);
		printf("Another one? (y/N) ");
	}	while (toupper(getche())=='Y');
	return(0);
}

boolean scanfidonumber(char *s, fidonumber *f)
{
	char fn[20];
	char nr[10];
	char s1,s2;
	int i,j;
	unsigned int n;
	boolean done=FALSE;

	f->zone=Aka[0].zone;
	f->net=Aka[0].net;
	f->node=Aka[0].node;
	f->point=0;

	i=0;
	while (s[i] && !isspace(s[i]) && i<sizeof(fn)-1) { fn[i]=s[i]; i++; }
	fn[i]=0;

	i=0;
	s2=s1=0;
	if (fn[0]=='.')
	{
		if (fn[1]==0) { f->point=Aka[0].point; return(TRUE); }
		if (!isdigit(fn[1])) return(FALSE);
		i++;
		s2='.';
	}

	while (!done)
	{
		s1=s2;
		j=0;
		while (isdigit(fn[i]) && j<sizeof(nr)-1) {nr[j]=fn[i]; i++; j++; }
		nr[j]=0;
		s2=fn[i];
		if (sscanf(nr,"%u",&n)!=1) return(FALSE);
		i++;
		done=(s2==0);

		switch (s1)
		{
			case 0	:	switch (s2)
							{
								case 0  :	f->node=n; break;
								case '.':	f->node=n; break;
								case '/':	f->net=n; break;
								case ':':	f->zone=n; break;
								default :	return(FALSE);
							}
							break;
			case '.'	:	if (s2!=0) return(FALSE);
							else f->point=n;
							break;
			case '/'	:	if (s2==0 || s2=='.') f->node=n;
							else return(FALSE);
							break;
			case ':'	:	if (s2=='/') f->net=n;
							else return(FALSE);
							break;
			default	:	return(FALSE);
		}
	}
	return(TRUE);
}


/* Crc - 32 BIT ANSI X3.66 CRC checksum files */

/**********************************************************************\
|* Demonstration program to compute the 32-bit CRC used as the frame  *|
|* check sequence in ADCCP (ANSI X3.66, also known as FIPS PUB 71     *|
|* and FED-STD-1003, the U.S. versions of CCITT's X.25 link-level     *|
|* protocol).  The 32-bit FCS was added via the Federal Register,     *|
|* 1 June 1982, p.23798.  I presume but don't know for certain that   *|
|* this polynomial is or will be included in CCITT V.41, which        *|
|* defines the 16-bit CRC (often called CRC-CCITT) polynomial.  FIPS  *|
|* PUB 78 says that the 32-bit FCS reduces otherwise undetected       *|
|* errors by a factor of 10^-5 over 16-bit FCS.                       *|
\**********************************************************************/

/* Copyright (C) 1986 Gary S. Brown.  You may use this program, or
   code or tables extracted from it, as desired without restriction.*/

/* First, the polynomial itself and its table of feedback terms.  The  */
/* polynomial is                                                       */
/* X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0 */
/* Note that we take it "backwards" and put the highest-order term in  */
/* the lowest-order bit.  The X^32 term is "implied"; the LSB is the   */
/* X^31 term, etc.  The X^0 term (usually shown as "+1") results in    */
/* the MSB being 1.                                                    */

/* Note that the usual hardware shift register implementation, which   */
/* is what we're using (we're merely optimizing it by doing eight-bit  */
/* chunks at a time) shifts bits into the lowest-order term.  In our   */
/* implementation, that means shifting towards the right.  Why do we   */
/* do it this way?  Because the calculated CRC must be transmitted in  */
/* order from highest-order term to lowest-order term.  UARTs transmit */
/* characters in order from LSB to MSB.  By storing the CRC this way,  */
/* we hand it to the UART in the order low-byte to high-byte; the UART */
/* sends each low-bit to hight-bit; and the result is transmission bit */
/* by bit from highest- to lowest-order term without requiring any bit */
/* shuffling on our part.  Reception works similarly.                  */

/* The feedback terms table consists of 256, 32-bit entries.  Notes:   */
/*                                                                     */
/*  1. The table can be generated at runtime if desired; code to do so */
/*     is shown later.  It might not be obvious, but the feedback      */
/*     terms simply represent the results of eight shift/xor opera-    */
/*     tions for all combinations of data and CRC register values.     */
/*                                                                     */
/*  2. The CRC accumulation logic is the same for all CRC polynomials, */
/*     be they sixteen or thirty-two bits wide.  You simply choose the */
/*     appropriate table.  Alternatively, because the table can be     */
/*     generated at runtime, you can start by generating the table for */
/*     the polynomial in question and use exactly the same "updcrc",   */
/*     if your application needn't simultaneously handle two CRC       */
/*     polynomials.  (Note, however, that XMODEM is strange.)          */
/*                                                                     */
/*  3. For 16-bit CRCs, the table entries need be only 16 bits wide;   */
/*     of course, 32-bit entries work OK if the high 16 bits are zero. */
/*                                                                     */
/*  4. The values must be right-shifted by eight bits by the "updcrc"  */
/*     logic; the shift must be unsigned (bring in zeroes).  On some   */
/*     hardware you could probably optimize the shift in assembler by  */
/*     using byte-swap instructions.                                   */

const unsigned long crc_32_tab[] = { /* CRC polynomial 0xedb88320 */
0x00000000L, 0x77073096L, 0xee0e612cL, 0x990951baL, 0x076dc419L, 0x706af48fL,
0xe963a535L, 0x9e6495a3L, 0x0edb8832L, 0x79dcb8a4L, 0xe0d5e91eL, 0x97d2d988L,
0x09b64c2bL, 0x7eb17cbdL, 0xe7b82d07L, 0x90bf1d91L, 0x1db71064L, 0x6ab020f2L,
0xf3b97148L, 0x84be41deL, 0x1adad47dL, 0x6ddde4ebL, 0xf4d4b551L, 0x83d385c7L,
0x136c9856L, 0x646ba8c0L, 0xfd62f97aL, 0x8a65c9ecL, 0x14015c4fL, 0x63066cd9L,
0xfa0f3d63L, 0x8d080df5L, 0x3b6e20c8L, 0x4c69105eL, 0xd56041e4L, 0xa2677172L,
0x3c03e4d1L, 0x4b04d447L, 0xd20d85fdL, 0xa50ab56bL, 0x35b5a8faL, 0x42b2986cL,
0xdbbbc9d6L, 0xacbcf940L, 0x32d86ce3L, 0x45df5c75L, 0xdcd60dcfL, 0xabd13d59L,
0x26d930acL, 0x51de003aL, 0xc8d75180L, 0xbfd06116L, 0x21b4f4b5L, 0x56b3c423L,
0xcfba9599L, 0xb8bda50fL, 0x2802b89eL, 0x5f058808L, 0xc60cd9b2L, 0xb10be924L,
0x2f6f7c87L, 0x58684c11L, 0xc1611dabL, 0xb6662d3dL, 0x76dc4190L, 0x01db7106L,
0x98d220bcL, 0xefd5102aL, 0x71b18589L, 0x06b6b51fL, 0x9fbfe4a5L, 0xe8b8d433L,
0x7807c9a2L, 0x0f00f934L, 0x9609a88eL, 0xe10e9818L, 0x7f6a0dbbL, 0x086d3d2dL,
0x91646c97L, 0xe6635c01L, 0x6b6b51f4L, 0x1c6c6162L, 0x856530d8L, 0xf262004eL,
0x6c0695edL, 0x1b01a57bL, 0x8208f4c1L, 0xf50fc457L, 0x65b0d9c6L, 0x12b7e950L,
0x8bbeb8eaL, 0xfcb9887cL, 0x62dd1ddfL, 0x15da2d49L, 0x8cd37cf3L, 0xfbd44c65L,
0x4db26158L, 0x3ab551ceL, 0xa3bc0074L, 0xd4bb30e2L, 0x4adfa541L, 0x3dd895d7L,
0xa4d1c46dL, 0xd3d6f4fbL, 0x4369e96aL, 0x346ed9fcL, 0xad678846L, 0xda60b8d0L,
0x44042d73L, 0x33031de5L, 0xaa0a4c5fL, 0xdd0d7cc9L, 0x5005713cL, 0x270241aaL,
0xbe0b1010L, 0xc90c2086L, 0x5768b525L, 0x206f85b3L, 0xb966d409L, 0xce61e49fL,
0x5edef90eL, 0x29d9c998L, 0xb0d09822L, 0xc7d7a8b4L, 0x59b33d17L, 0x2eb40d81L,
0xb7bd5c3bL, 0xc0ba6cadL, 0xedb88320L, 0x9abfb3b6L, 0x03b6e20cL, 0x74b1d29aL,
0xead54739L, 0x9dd277afL, 0x04db2615L, 0x73dc1683L, 0xe3630b12L, 0x94643b84L,
0x0d6d6a3eL, 0x7a6a5aa8L, 0xe40ecf0bL, 0x9309ff9dL, 0x0a00ae27L, 0x7d079eb1L,
0xf00f9344L, 0x8708a3d2L, 0x1e01f268L, 0x6906c2feL, 0xf762575dL, 0x806567cbL,
0x196c3671L, 0x6e6b06e7L, 0xfed41b76L, 0x89d32be0L, 0x10da7a5aL, 0x67dd4accL,
0xf9b9df6fL, 0x8ebeeff9L, 0x17b7be43L, 0x60b08ed5L, 0xd6d6a3e8L, 0xa1d1937eL,
0x38d8c2c4L, 0x4fdff252L, 0xd1bb67f1L, 0xa6bc5767L, 0x3fb506ddL, 0x48b2364bL,
0xd80d2bdaL, 0xaf0a1b4cL, 0x36034af6L, 0x41047a60L, 0xdf60efc3L, 0xa867df55L,
0x316e8eefL, 0x4669be79L, 0xcb61b38cL, 0xbc66831aL, 0x256fd2a0L, 0x5268e236L,
0xcc0c7795L, 0xbb0b4703L, 0x220216b9L, 0x5505262fL, 0xc5ba3bbeL, 0xb2bd0b28L,
0x2bb45a92L, 0x5cb36a04L, 0xc2d7ffa7L, 0xb5d0cf31L, 0x2cd99e8bL, 0x5bdeae1dL,
0x9b64c2b0L, 0xec63f226L, 0x756aa39cL, 0x026d930aL, 0x9c0906a9L, 0xeb0e363fL,
0x72076785L, 0x05005713L, 0x95bf4a82L, 0xe2b87a14L, 0x7bb12baeL, 0x0cb61b38L,
0x92d28e9bL, 0xe5d5be0dL, 0x7cdcefb7L, 0x0bdbdf21L, 0x86d3d2d4L, 0xf1d4e242L,
0x68ddb3f8L, 0x1fda836eL, 0x81be16cdL, 0xf6b9265bL, 0x6fb077e1L, 0x18b74777L,
0x88085ae6L, 0xff0f6a70L, 0x66063bcaL, 0x11010b5cL, 0x8f659effL, 0xf862ae69L,
0x616bffd3L, 0x166ccf45L, 0xa00ae278L, 0xd70dd2eeL, 0x4e048354L, 0x3903b3c2L,
0xa7672661L, 0xd06016f7L, 0x4969474dL, 0x3e6e77dbL, 0xaed16a4aL, 0xd9d65adcL,
0x40df0b66L, 0x37d83bf0L, 0xa9bcae53L, 0xdebb9ec5L, 0x47b2cf7fL, 0x30b5ffe9L,
0xbdbdf21cL, 0xcabac28aL, 0x53b39330L, 0x24b4a3a6L, 0xbad03605L, 0xcdd70693L,
0x54de5729L, 0x23d967bfL, 0xb3667a2eL, 0xc4614ab8L, 0x5d681b02L, 0x2a6f2b94L,
0xb40bbe37L, 0xc30c8ea1L, 0x5a05df1bL, 0x2d02ef8dL };

#define UPDC32(octet, crc) (crc_32_tab[(int)(((crc)^(octet))&0xff)]^((crc)>>8))

unsigned long calc_crc32(char *name)
{
		unsigned long crc32;
		int c,i=0;

		crc32 = 0xFFFFFFFFL;

		while ((c=name[i])!=0)
      {
			crc32 = UPDC32(c, crc32);
			i++;
      }
		return(crc32);
}

int readconfig(string config, char *p)
{
	FILE *conf;
	int akacnt=0,linkcnt=0,tagcnt=0,argp,i,r,begpnt,k,l,m;
	char key[11];
	string line;
	char drive[MAXPATH];
	char *s;

	static char keywords[][11]={"AKA","LOG","INBOUND","SECURE","OUTBOUND",
										"PACKETS","PKTTEMP","NODE","HISFILE","MYFILE",
										"RETAG","END","RETEAR","REPATH","STRIPS",
										"TOSSER","VERBOSE","VERBOSITY","PRODCODE"};
	static char tossers[][11]={TOSSER0,TOSSER1,TOSSER2};

	fnsplit(p,drive,line,NULL,NULL);
	strcat(drive,line);
	strcat(drive,config);
	s=config;

	if ((conf=fopen(config,"rt"))==NULL &&
							 (s=drive,conf=fopen(drive,"rt"))==NULL) return(NOCONFIG);
	printf("Reading %s\n",s);

	while (fgets(line,sizeof(line)-1,conf)!=NULL)
	{
		clinenr++;
		begpnt=0;
		while (isspace(line[begpnt])) begpnt++;
		if (line[begpnt]==0 || line[begpnt]==';') continue;

		argp=begpnt;
		while (line[argp]!='=' && !isspace(line[argp]) && line[argp]) argp++;
		argp++;
		l=min(argp-begpnt-1,sizeof(keywords[0]));
		while (isspace(line[argp]) && line[argp]!='\n' && line[argp]) argp++;
		for (i=0;i<l;i++)  key[i]=toupper(line[begpnt+i]);
		key[i]=0;
		k=strlen(key);
		for (i=0;i<sizeof(keywords)/sizeof(keywords[0]);i++)
		{
			l=max(3,(min(k,sizeof(keywords[0]))));
			if (strncmp(key,keywords[i],l)==0) break;
		}

		switch (i)
		{
			case 0:  if (akacnt<MAXAKAS)						/* AKA */
						{
							if (scanfidonumber(&line[argp],&Aka[akacnt])) akacnt++;
							else return(WRONGNUMBER);
						}
						else return(TOOMANYAKAS);
						break;
			case 1:	l=argp; m=0;                        /* LOG */
						while (!isspace(line[l]) && line[l])
						{
							Logfile[m]=toupper(line[l]);
							l++; m++;
						}
						Logfile[m]=0;
						removelf(Logfile);
						if (strlen(Logfile)==0) return(NAMEMISSING);
						break;
			case 2:	copydirname(Inbound,&line[argp]);   /* INBOUND */
						if (strlen(Inbound)==0) return(NAMEMISSING);
						break;
			case 3:	copydirname(Secure,&line[argp]);    /* SECURE */
						if (strlen(Secure)==0) return(NAMEMISSING);
						break;
			case 4:	copydirname(Outbound,&line[argp]);  /* OUTBOUND */
						if (strlen(Outbound)==0) return(NAMEMISSING);
						break;
			case 5:	copydirname(Packets,&line[argp]);   /* PACKETS */
						if (strlen(Packets)==0) return(NAMEMISSING);
						break;
			case 6:	copydirname(Pkttemp,&line[argp]);   /* PKTTEMP */
						if (strlen(Pkttemp)==0) return(NAMEMISSING);
         			break;
			case 7:  if (linkcnt<MAXLINKS)               /* NODE */
						{
							scanfidonumber(&line[argp],&links[linkcnt]);
							Tagpntrs[linkcnt]=tagcnt;
							if (linkcnt) Flags[linkcnt]=Flags[linkcnt-1];
							else
							{
								if (retear) Flags[linkcnt]|=RETEAR;
								if (repath) Flags[linkcnt]|=REPATH;
								if (stripseenby) Flags[linkcnt]|=STRIPS;
							}
							linkcnt++;
						}
						else return(TOOMANYLINKS);
						break;
			case 8:  if (linkcnt<MAXLINKS)               /* HISFILE */
						{
							strncpy(Hisfiles[linkcnt-1],&line[argp],8);
							Hisfiles[linkcnt-1][8]=0;
							removelf(Hisfiles[linkcnt-1]);
							for (l=strlen(Hisfiles[linkcnt-1])-1;l>=0 && Hisfiles[linkcnt-1][l]==' ';l--)
									Hisfiles[linkcnt-1][l]=0;
							if (strlen(Hisfiles[linkcnt-1])==0) return(NAMEMISSING);
						}
						else return(TOOMANYLINKS);
						break;
			case 9:  if (linkcnt<MAXLINKS)               /* MYFILE */
						{
							strncpy(Myfiles[linkcnt-1],&line[argp],8);
							Myfiles[linkcnt-1][8]=0;
							removelf(Myfiles[linkcnt-1]);
							for (l=strlen(Myfiles[linkcnt-1])-1;l>=0 && Myfiles[linkcnt-1][l]==' ';l--)
									Myfiles[linkcnt-1][l]=0;
							if (strlen(Myfiles[linkcnt-1])==0) return(NAMEMISSING);
						}
						else return(TOOMANYLINKS);
						break;
			case 10:	if (linkcnt<MAXLINKS && tagcnt<MAXTAGS) /* RETAG */
						{
							l=argp; m=0;
							while (!isspace(line[l]))
							{
								Histags[tagcnt][m]=toupper(line[l]);
								l++; m++;
							}
							if (m==0) return(AREAMISSING);
							l++; m=0;
							while (isspace(line[l]) && line[l]!='\n' && line[l]) l++;
							while (!isspace(line[l]) && line[l])
							{
								Mytags[tagcnt][m]=toupper(line[l]);
								l++; m++;
							}
							if (m==0) return(AREAMISSING);
							if (linkcnt) Tagaantal[linkcnt-1]++;
							tagcnt++;
						}
						else return(TOOMANYTAGS);
						break;
			case 11:	return(0);                              /* END */
			case 12: if (linkcnt)                            /* RETEAR */
						{
							if (strnicmp(&line[argp],"ON",2)==0)
								{
									Flags[linkcnt-1]|=RETEAR;
								}
							else
							{
								if (strnicmp(&line[argp],"OFF",3)==0)
								{
									Flags[linkcnt-1]&=(~RETEAR);
								}
								else return(ONOROFF);
							}
						}
						else
						{
							if (strnicmp(&line[argp],"ON",2)==0) retear=TRUE;
							else
							{
								if (strnicmp(&line[argp],"OFF",3)==0) retear=FALSE;
								else return(ONOROFF);
							}
						}
						break;
			case 13:	if (linkcnt)                            /* REPATH */
						{
							if (strnicmp(&line[argp],"ON",2)==0)
								{
									Flags[linkcnt-1]|=REPATH;
								}
							else
							{
								if (strnicmp(&line[argp],"OFF",3)==0)
								{
									Flags[linkcnt-1]&=(~REPATH);
								}
								else return(ONOROFF);
							}
						}
						else
						{
							if (strnicmp(&line[argp],"ON",2)==0) repath=TRUE;
							else
							{
								if (strnicmp(&line[argp],"OFF",3)==0) repath=FALSE;
								else return(ONOROFF);
							}
						}
						break;
			case 14:	if (linkcnt)                            /* STRIPS */
						{
							if (strnicmp(&line[argp],"ON",2)==0)
								{
									Flags[linkcnt-1]|=STRIPS;
								}
							else
							{
								if (strnicmp(&line[argp],"OFF",3)==0)
								{
									Flags[linkcnt-1]&=(~STRIPS);
								}
								else return(ONOROFF);
							}
						}
						else
						{
							if (strnicmp(&line[argp],"ON",2)==0) stripseenby=TRUE;
							else
							{
								if (strnicmp(&line[argp],"OFF",3)==0) stripseenby=FALSE;
								else return(ONOROFF);
							}
						}
						break;
			case 15:	for (i=0;i<(sizeof(tossers)/sizeof(tossers[0]));i++)
						{
							if (strnicmp(&line[argp],tossers[i],4)==0)
							{
								Tosserid=i;
								r=readtosserconfig(Tosserid,&akacnt);
								if (r) return(r);
								break;
							}
						}
						if (i>=(sizeof(tossers)/sizeof(tossers[0]))) return(UNKNOWNTOSSER);
						break;
			case 16:						/* VERBOSE & VERBOSITY */
			case 17:	if (strnicmp(&line[argp],"ON",2)==0) Verbose=1;
						else
						{
							if (strnicmp(&line[argp],"OFF",3)==0) Verbose=0;
							else
							{
								sscanf(&line[argp],"%d",&Verbose);
							}
						}
						break;
			case 18:	if (strnicmp(&line[argp],"ON",2)==0) FtscprodcodeOn=1;
						else
						{
							if (strnicmp(&line[argp],"OFF",3)==0) FtscprodcodeOn=0;
							else
							{
								sscanf(&line[argp],"%d",&FtscprodcodeOn);
							}
						}
						break;
			default: strcpy(logline,key);
						return(BADKEYWORD);
		}
	}
	return(0);
}

void copydirname(char *s, char *t)
{
	int i=0;
	while (!isspace(t[i]) && i<sizeof(string)-1) { s[i]=toupper(t[i]); i++; }
	removelf(s);
	addslash(s);
}


void removelf(char *s)
{
	int i=0;

	do
	{
		if (s[i]=='\n') {s[i]=0; return; }
		i++;
	}	while (s[i]);
}

void addslash(char *s)
{
	int i;

	i=strlen(s)-1;

	if (s[i]!='\\') strcat(s,"\\");
}

void removeslash(char *s)
{
	int i;

	i=strlen(s)-1;

	if (s[i]=='\\') s[i]=0;
}


int isaka(unsigned int z, unsigned int n, unsigned int f, unsigned int p)
{
	int i;

	for (i=0;i<MAXAKAS && Aka[i].zone;i++)
	{
		if (z==Aka[i].zone && n==Aka[i].net && f==Aka[i].node && p==Aka[i].point)
		return(i);
	}
	return(-1);
}

void report_error(int err)
{
	printf("\nError %d reading config file line %d: ",err,clinenr);

	switch (err)
	{
		case TOOMANYAKAS:		printf("Too many aka's."); break;
		case TOOMANYLINKS:	printf("Too many links."); break;
		case TOOMANYTAGS:		printf("Too many tags."); break;
		case WRONGNUMBER:		printf("Not a proper nodenumber."); break;
		case ONOROFF:			printf("Argument must be 'ON'or 'OFF'."); break;
		case BADKEYWORD:		printf("Unknown keyword '%s'.",logline); break;
		case NAMEMISSING:		printf("Missing filename or path."); break;
		case AREAMISSING:		printf("Missing areaname."); break;
		case UNKNOWNTOSSER:	printf("Unknown tosser."); break;
		case NOGEENVIRON:		printf("GE environment variable not set."); break;
		case NOFMENVIRON:		printf("FMAIL environment variable not set."); break;
		case NOGESETUP:		printf("Can't open SETUP.GE"); break;
		case NOFMCONFIG:		printf("Can't open FMAIL.CFG"); break;
		case GESETUPERR:		printf("Error reading SETUP.GE"); break;
		case FMCFGERR:			printf("Error reading FMAIL.CFG"); break;
		case NOCONFIG:			printf("Can not open config file."); break;
		case MEMALLOCERR:		printf("Memory allocation error.");break;
		default:					printf("Unknown error."); break;
	}
	printf("\n\a");
	printf("%s aborted\n",PROGNAME);
}

void renameleftpackets(char *dir)
{
	int i,n;
	string pktname, pktname2;

	n=findpackets(dir);
	{
		for (i=0;i<n;i++)
		{
			strcpy(pktname,dir);						strcpy(pktname2,dir);
			strcat(pktname,pkts_tounpack[i]);   strcat(pktname2,pkts_tounpack[i]);
			strcat(pktname,".PKT");             strcat(pktname2,".$$$");
			rename(pktname,pktname2);
		}
	}
}

/* create temp dirs if the do not exist */

boolean makedirs(void)
{
	struct ffblk infoblk;
	int err1=0,err2=0,err3=0;

	strcpy(Pkttemp_i,Pkttemp);
	strcat(Pkttemp_i,TEMP_I);
	strcpy(Pkttemp_o,Pkttemp);
	strcat(Pkttemp_o,TEMP_O);
	removeslash(Pkttemp);

	if (findfirst(Pkttemp,&infoblk,FA_DIREC)!=0)
	{
	  if ((err1=mkdir(Pkttemp))==0)
		{
			sprintf(logline,"Created directory: %s\n",Pkttemp);
			pstring(logline,0);
		}
	}

	if (err1==0)
	{
		if (findfirst(Pkttemp_i,&infoblk,FA_DIREC)!=0)
		{
			if ((err2=mkdir(Pkttemp_i))==0)
			{
				sprintf(logline,"Created directory: %s\n",Pkttemp_i);
				pstring(logline,0);
			}
		}

		if (findfirst(Pkttemp_o,&infoblk,FA_DIREC)!=0)
		{
			if ((err3=mkdir(Pkttemp_o))==0)
			{
				sprintf(logline,"Created directory: %s\n",Pkttemp_o);
				pstring(logline,0);
			}
		}
	}
	if (err1==0 && err2==0 && err3==0)
	{
		addslash(Pkttemp);
		addslash(Pkttemp_i);
		addslash(Pkttemp_o);
		return(TRUE);
	}
	else
	{
		if (err1)
		{
			sprintf(logline,"Error creating directory: %s\n",Pkttemp);
			pstring(logline,0);
			return(FALSE);
		}
		else
		{
			if (err2)
			{
				sprintf(logline,"Error creating directory: %s\n",Pkttemp_i);
				pstring(logline,0);
			}
			if (err3)
			{
				sprintf(logline,"Error creating directory: %s\n",Pkttemp_o);
				pstring(logline,0);
			}
		}
		return(FALSE);
	}
}

int readtosserconfig(int tossertype, int *akacnt)
{
	if (tossertype!=GECHO && tossertype!=FMAIL ) return(UNKNOWNTOSSER);
	if (tossertype==GECHO) return(readgesetup(akacnt));
	return(readfmailcfg(akacnt));
}

int readgesetup(int *akacnt)
{
	int i,gshandle,result;
	char *ge;
	string gesetupname;
	SETUP_GE far *ge_setup;

	ge_setup=farmalloc(sizeof(SETUP_GE));

	if (ge_setup==NULL) return(MEMALLOCERR);

	ge=getenv("GE");
	if (ge==NULL) {farfree(ge_setup); return(NOGEENVIRON); }
	i=0;
	while(ge[i] && ge[i]!=' ') gesetupname[i]=ge[i++];
	gesetupname[i]=0;
	addslash(gesetupname);
	strcat(gesetupname,"SETUP.GE");
	printf("Reading %s\n",gesetupname);

	gshandle=open(gesetupname,O_RDONLY|O_DENYWRITE|O_BINARY);
	if (gshandle==-1) { farfree(ge_setup); return(NOGESETUP); }

	result=read(gshandle,ge_setup,sizeof(SETUP_GE));
	close(gshandle);
	if (result<0x1342) { farfree(ge_setup); return(GESETUPERR); }
	if (ge_setup->sysrev!=GE_THISREV) {farfree(ge_setup); return(GESETUPERR);}

	for (i=0;i<AKAS;i++)
	{
		if (*akacnt>=MAXAKAS) break;
		if (ge_setup->aka[i].zone==0) break;
		Aka[*akacnt].zone=ge_setup->aka[i].zone;
		Aka[*akacnt].net=ge_setup->aka[i].net;
		Aka[*akacnt].node=ge_setup->aka[i].node;
		Aka[*akacnt].point=ge_setup->aka[i].point;
		(*akacnt)++;
	}
	strcpy(Inbound,ge_setup->inbound_path);
	strcpy(Outbound,ge_setup->outbound_path);
	strcpy(Secure,ge_setup->secure_path);
	if (result>=sizeof(SETUP_GE)) strcpy(Packets,ge_setup->inpkts_path);
	else strcpy(Packets,ge_setup->inbound_path);
	strcpy(Logfile,ge_setup->logfile);

	farfree(ge_setup);

	return(0);
}

int readfmailcfg(int *akacnt)
{
#define OLDSIZE 8192

	int i,fmhandle,result;
	char *fm;
	string fmcfgname;
	configType far *fmail_cfg;

	fmail_cfg=farmalloc(sizeof(configType));

	if (fmail_cfg==NULL) return(MEMALLOCERR);

	fm=getenv("FMAIL");
	if (fm==NULL) {farfree(fmail_cfg); return(NOFMENVIRON); }
	i=0;
	while(fm[i] && fm[i]!=' ') fmcfgname[i]=fm[i++];
	fmcfgname[i]=0;
	addslash(fmcfgname);
	strcat(fmcfgname,"FMAIL.CFG");
	printf("Reading %s\n",fmcfgname);
	fmhandle=open(fmcfgname,O_RDONLY|O_DENYWRITE|O_BINARY);
	if (fmhandle==-1) {farfree(fmail_cfg); return(NOFMCONFIG); }

	result=read(fmhandle,fmail_cfg,sizeof(configType));
	close(fmhandle);
	Fm10x=(result==OLDSIZE);
	if (!Fm10x && result<sizeof(configType)) {farfree(fmail_cfg); return(FMCFGERR); }

	for (i=0;i<AKAS;i++)
	{
		if (*akacnt>=MAXAKAS) break;
		if (Fm10x)
		{
			if (fmail_cfg->_akaList[i].nodeNum.zone==0) break;
			Aka[*akacnt].zone=fmail_cfg->_akaList[i].nodeNum.zone;
			Aka[*akacnt].net=fmail_cfg->_akaList[i].nodeNum.net;
			Aka[*akacnt].node=fmail_cfg->_akaList[i].nodeNum.node;
			Aka[*akacnt].point=fmail_cfg->_akaList[i].nodeNum.point;
		}
		else
		{
			if (fmail_cfg->akaList[i].nodeNum.zone==0) break;
			Aka[*akacnt].zone=fmail_cfg->akaList[i].nodeNum.zone;
			Aka[*akacnt].net=fmail_cfg->akaList[i].nodeNum.net;
			Aka[*akacnt].node=fmail_cfg->akaList[i].nodeNum.node;
			Aka[*akacnt].point=fmail_cfg->akaList[i].nodeNum.point;
		}
		(*akacnt)++;
	}
#pragma warn -ucp			/* to avoid warning about mixed signs */
	strcpy(Inbound,fmail_cfg->inPath);
	strcpy(Outbound,fmail_cfg->outPath);
	strcpy(Secure,fmail_cfg->securePath);
	strcpy(Packets,fmail_cfg->inPath);
	strcpy(Logfile,fmail_cfg->logName);
#pragma warn +ucp

	farfree(fmail_cfg);

	return(0);
}



/* That's it folks */

